import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange


class Quantizer(nn.Module):
    def __init__(self, entry_num, entry_dim):
        super().__init__()

        self.entry_num = entry_num
        self.entry_dim = entry_dim
        self.decay = 0.9
        self.entry = nn.Embedding(self.entry_num, self.entry_dim)
        self.register_buffer("entry_prob", torch.zeros(self.entry_num))

    def init_codebook(self, z, method):
        if method == "Random":
            self.entry.weight.data.uniform_(-1.0 / self.entry_num, 1.0 / self.entry_num)
        elif method == "Kmeans":
            import faiss

            d = z.shape[1]
            kmeans = faiss.Kmeans(d, self.entry_num, spherical=True, gpu=True)
            kmeans.train(z)
            D, I = kmeans.index.search(z, 1)
            assignments = I.reshape(-1)
            centers = np.zeros((self.entry_num, d))
            for i in range(self.entry_num):
                centers[i] = z[assignments == i].mean(axis=0)
            self.entry.weight.data.copy_(torch.from_numpy(centers))
        elif method == "Geometric":
            from geosketch import gs

            sketch_index = gs(z, self.entry_num, replace=False)
            self.entry.weight.data.copy_(torch.from_numpy(z[sketch_index]))

    def forward(self, e, return_assignment):
        # cosine similarity
        normed_e = F.normalize(e, dim=1).detach()
        normed_c = F.normalize(self.entry.weight, dim=1)
        sim = torch.einsum("bd,dn->bn", normed_e, rearrange(normed_c, "n d -> d n"))

        # entry assignment
        assignment_indices = torch.argmax(sim, dim=1)
        assignments = torch.zeros(
            assignment_indices.unsqueeze(1).shape[0], self.entry_num, device=e.device
        )
        assignments.scatter_(1, assignment_indices.unsqueeze(1), 1)
        avg_probs = torch.mean(assignments, dim=0)

        # quantize
        e_q = torch.matmul(assignments, self.entry.weight)
        # L_C
        loss = torch.mean((e_q - e.detach()) ** 2)

        if self.training:
            # update the entry usage
            self.entry_prob.mul_(self.decay).add_(avg_probs, alpha=1 - self.decay)

            # deal with small entries
            norm_distance = F.softmax(1 - sim, dim=1)
            norm_distance = torch.max(norm_distance, dim=1).values
            dis_indices = torch.multinomial(
                norm_distance, num_samples=self.entry_num, replacement=True
            ).view(-1)
            random_feat = e.detach()[dis_indices]
            beta_s = (
                torch.exp(-self.entry_prob * self.entry_num * 100 - 1e-3)
                .unsqueeze(1)
                .repeat(1, self.entry_dim)
            )
            self.entry.weight.data = (
                self.entry.weight.data * (1 - beta_s) + random_feat * beta_s
            )

            # deal with large entries
            if self.entry_prob.sum() + 1e-4 >= 1:
                sim_t = sim.t()
                median_distance = torch.median(sim_t, dim=1).values
                median_distance = torch.abs(sim_t - median_distance[:, None])
                dis_indices = torch.multinomial(
                    F.softmax(-median_distance, dim=1), num_samples=1
                ).view(-1)
                random_feat = e.detach()[dis_indices]
                beta_l = (
                    torch.exp(-self.entry_prob.mean() / self.entry_prob * 10 - 1e-3)
                    .unsqueeze(1)
                    .repeat(1, self.entry_dim)
                )
                self.entry.weight.data = (
                    self.entry.weight.data * (1 - beta_l) + random_feat * beta_l
                )

        if return_assignment:
            top2_sim = torch.topk(sim, 2, dim=1).values
            delta_conf = top2_sim[:, 0] - top2_sim[:, 1]
            loss_c_sum = torch.sum((e_q - e.detach()) ** 2)

            return assignment_indices, delta_conf, loss_c_sum
        else:
            return e_q, loss


class Encoder(nn.Module):
    def __init__(self, input_dim, entry_dim):
        super().__init__()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, 512),
            nn.BatchNorm1d(512),
            nn.ReLU(),
            nn.Linear(512, 128),
            nn.BatchNorm1d(128),
            nn.ReLU(),
            nn.Linear(128, entry_dim),
        )

    def forward(self, input):
        return self.encoder(input)


class Decoder(nn.Module):
    def __init__(self, entry_dim, output_dim):
        super().__init__()
        self.docoder = nn.Sequential(
            nn.Linear(entry_dim, 128),
            nn.BatchNorm1d(128),
            nn.ReLU(),
            nn.Linear(128, 512),
            nn.BatchNorm1d(512),
            nn.ReLU(),
        )
        self.decoder_mean = nn.Linear(512, output_dim)
        self.decoder_disp = nn.Sequential(
            nn.Linear(512, output_dim),
            nn.Softplus(),
        )

    def forward(self, input):
        decode = self.docoder(input)
        mean = torch.clamp(torch.exp(self.decoder_mean(decode)), 1e-5, 1e6)
        disp = torch.clamp(self.decoder_disp(decode), 1e-4, 1e4)
        return mean, disp


class Decoder_ATAC(nn.Module):
    def __init__(self, entry_dim, output_dim):
        super().__init__()
        self.docoder = nn.Sequential(
            nn.BatchNorm1d(entry_dim),
            nn.Linear(entry_dim, output_dim),
        )

    def forward(self, input):
        mean = torch.clamp(torch.exp(self.docoder(input)), 1e-5, 1e6)
        return mean, None


def get_decoder(data_type, entry_dim, output_dim):
    if data_type == "RNA" or data_type == "ADT":
        return Decoder(entry_dim, output_dim)
    elif data_type == "ATAC":
        return Decoder_ATAC(entry_dim, output_dim)


class MetaQ(nn.Module):
    def __init__(self, input_dims, data_types, entry_num, entry_dim=32):
        super(MetaQ, self).__init__()
        self.omics_num = len(input_dims)
        self.encoders = nn.ModuleList(
            [Encoder(input_dim, entry_dim) for input_dim in input_dims]
        )
        self.quantizer = Quantizer(entry_num, entry_dim * self.omics_num)
        self.decoders = nn.ModuleList(
            [
                get_decoder(data_type, entry_dim, input_dim)
                for input_dim, data_type in zip(input_dims, data_types)
            ]
        )
        self.decoders_q = nn.ModuleList(
            [
                get_decoder(data_type, entry_dim * self.omics_num, input_dim)
                for input_dim, data_type in zip(input_dims, data_types)
            ]
        )

    def copy_decoder_q(self):
        if self.omics_num == 1:
            self.decoders_q[0].load_state_dict(self.decoders[0].state_dict())

    def quantize(self, hiddens, return_assignment=False):
        if self.omics_num > 1:
            hidden = torch.cat(hiddens, dim=1)
        else:
            hidden = hiddens[0]
        return self.quantizer(hidden, return_assignment)

    def forward(self, inputs):
        hiddens = []
        for i in range(self.omics_num):
            hiddens.append(self.encoders[i](inputs[i]))
        return hiddens

    def decode(self, hiddens):
        means, disps = [], []
        for i in range(self.omics_num):
            mean, disp = self.decoders[i](hiddens[i])
            means.append(mean)
            disps.append(disp)
        return means, disps

    def decode_q(self, hidden):
        means, disps = [], []
        for i in range(self.omics_num):
            mean, disp = self.decoders_q[i](hidden)
            means.append(mean)
            disps.append(disp)
        return means, disps


def negative_binomial_loss(mean, disp, scale_factor, x):
    eps = 1e-12
    mean = mean * scale_factor

    t1 = torch.lgamma(disp + eps) + torch.lgamma(x + 1.0) - torch.lgamma(x + disp + eps)
    t2 = (disp + x) * torch.log(1.0 + (mean / (disp + eps))) + (
        x * (torch.log(disp + eps) - torch.log(mean + eps))
    )
    nb_final = t1 + t2
    result = nb_final

    return result


def poisson_loss(mean, scale_factor, x):
    eps = 1e-12
    mean = mean * scale_factor

    poisson = mean - x * torch.log(mean + eps) + torch.lgamma(x + 1)

    return poisson


def reconstruction_loss(mean, disp, scale_factor, x, data_type, reduction="mean"):
    if data_type == "RNA" or data_type == "ADT":
        loss = negative_binomial_loss(mean, disp, scale_factor, x)
    elif data_type == "ATAC":
        loss = poisson_loss(mean, scale_factor, x)
    if reduction == "mean":
        loss = torch.mean(loss)
    elif reduction == "sum":
        loss = torch.sum(loss)
    else:
        raise NotImplementedError
    return loss



import torch
import numpy as np
from .model import reconstruction_loss


def train_one_epoch(
    model,
    data_types,
    dataloader,
    optimizer,
    epoch,
    device,
):
    model.train(True)
    optimizer.zero_grad()

    omics_num = model.omics_num

    loss_rec_epoch = [0 for _ in range(omics_num)]
    loss_rec_q_epoch = [0 for _ in range(omics_num)]
    loss_c_epoch = 0

    for data in dataloader:
        x_list = []
        sf_list = []
        raw_list = []
        for i in range(omics_num):
            x_list.append(data["x"][i].to(device, non_blocking=True))
            sf_list.append(data["sf"][i].to(device, non_blocking=True))
            raw_list.append(data["raw"][i].to(device, non_blocking=True))

        hiddens = model(x_list)
        hidden_q, loss_c = model.quantize(hiddens)
        means, disps = model.decode(hiddens)
        means_q, disps_q = model.decode_q(hidden_q)

        loss_rec_all = 0
        loss_rec_q_all = 0
        for i in range(omics_num):
            loss_rec = reconstruction_loss(
                means[i], disps[i], sf_list[i], raw_list[i], data_types[i]
            )
            loss_rec_q = reconstruction_loss(
                means_q[i], disps_q[i], sf_list[i], raw_list[i], data_types[i]
            )
            loss_rec_all += loss_rec
            loss_rec_q_all += loss_rec_q
            loss_rec_epoch[i] += loss_rec.item()
            loss_rec_q_epoch[i] += loss_rec_q.item()
        loss_c_epoch += loss_c.item()

        loss = loss_rec_all + loss_rec_q_all + loss_c

        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

    if (epoch + 1) % 20 == 0 or epoch == 0:
        print("[Epoch %d]" % (epoch + 1), end=" ")
        for i in range(omics_num):
            if i > 0:
                print("|", end=" ")
            print(
                data_types[i]
                + ": Loss Rec=%.4f" % (loss_rec_epoch[i] / len(dataloader)),
                end=" ",
            )
            print(
                "Loss Rec Q=%.4f" % (loss_rec_q_epoch[i] / len(dataloader)),
                end=" ",
            )
        print("| Codebook: Loss C=%.4f" % (loss_c / len(dataloader)))

    loss_rec_q_epoch = np.array(loss_rec_q_epoch).mean() / len(dataloader)
    loss_c_epoch = loss_c / len(dataloader)

    return loss_rec_q_epoch, loss_c_epoch


def warm_one_epoch(
    model,
    data_types,
    dataloader,
    optimizer,
    epoch,
    device,
):
    model.train(True)
    optimizer.zero_grad()

    omics_num = model.omics_num

    loss_rec_epoch = [0 for _ in range(omics_num)]

    for data in dataloader:
        x_list = []
        sf_list = []
        raw_list = []
        for i in range(omics_num):
            x_list.append(data["x"][i].to(device, non_blocking=True))
            sf_list.append(data["sf"][i].to(device, non_blocking=True))
            raw_list.append(data["raw"][i].to(device, non_blocking=True))

        hiddens = model(x_list)
        means, disps = model.decode(hiddens)

        loss_rec_all = 0
        for i in range(omics_num):
            loss_rec = reconstruction_loss(
                means[i], disps[i], sf_list[i], raw_list[i], data_types[i]
            )
            loss_rec_all += loss_rec
            loss_rec_epoch[i] += loss_rec.item()

        loss = loss_rec_all

        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

    if (epoch + 1) % 20 == 0 or epoch == 0:
        print("[Epoch %d]" % (epoch + 1), end=" ")
        for i in range(omics_num):
            if i > 0:
                print("|", end=" ")
            print(
                data_types[i]
                + ": Loss Rec=%.4f" % (loss_rec_epoch[i] / len(dataloader)),
                end=" ",
            )
        print("")

    loss_rec_epoch = np.array(loss_rec_epoch).mean() / len(dataloader)

    return loss_rec_epoch


@torch.no_grad()
def inference(model, data_types, data_loader, device):
    model.eval()

    omics_num = model.omics_num

    embeds = []
    ids = []
    delta_confs = []
    loss_rec_all = 0
    loss_rec_q_all = 0
    loss_c_all = 0

    for data in data_loader:
        x_list = []
        sf_list = []
        raw_list = []
        for i in range(omics_num):
            x_list.append(data["x"][i].to(device, non_blocking=True))
            sf_list.append(data["sf"][i].to(device, non_blocking=True))
            raw_list.append(data["raw"][i].to(device, non_blocking=True))

        with torch.no_grad():
            hiddens = model(x_list)

            id, delta_conf, loss_c = model.quantize(hiddens, return_assignment=True)
            loss_c_all += loss_c

            hidden_q, _ = model.quantize(hiddens)
            means, disps = model.decode(hiddens)
            means_q, disps_q = model.decode_q(hidden_q)

            for i in range(omics_num):
                loss_rec = reconstruction_loss(
                    means[i],
                    disps[i],
                    sf_list[i],
                    raw_list[i],
                    data_types[i],
                    reduction="sum",
                )
                loss_rec_all += loss_rec
                loss_rec_q = reconstruction_loss(
                    means_q[i],
                    disps_q[i],
                    sf_list[i],
                    raw_list[i],
                    data_types[i],
                    reduction="sum",
                )
                loss_rec_q_all += loss_rec_q

        if omics_num > 1:
            hidden = torch.cat(hiddens, dim=1)
        else:
            hidden = hiddens[0]

        embeds.append(hidden.detach().cpu().numpy())
        ids.append(id.detach().cpu().numpy())
        delta_confs.append(delta_conf.detach().cpu().numpy())

    embeds = np.concatenate(embeds, axis=0)
    ids = np.concatenate(ids, axis=0)
    delta_confs = np.concatenate(delta_confs, axis=0)

    rec_q_percent = (loss_rec_all / loss_rec_q_all).item()
    loss_c_all = loss_c_all.item() / (
        data_loader.dataset.__len__() * model.quantizer.entry_dim
    )

    return embeds, ids, delta_confs, rec_q_percent, loss_c_all



import warnings

warnings.filterwarnings("ignore")

import os
import torch
import random
import argparse
import matplotlib
import numpy as np
import scanpy as sc
import seaborn as sns
from .model import MetaQ
from matplotlib import rcParams
from alive_progress import alive_bar
from .engine import train_one_epoch, warm_one_epoch, inference
from .data_utils import load_data, compute_metacell
from .eval_utils import (
    plot_metacell_umap,
    plot_metacell_size,
    plot_celltype_purity,
    plot_compactness_separation,
)


def main(args):
    device = torch.device(args.device)

    adata_list, dataloader_train, dataloader_eval, input_dims = load_data(args)
    omics_num = len(adata_list)

    print("Target metacell number:", args.metacell_num)

    net = MetaQ(
        input_dims=input_dims,
        data_types=args.data_type,
        entry_num=args.metacell_num,
    ).to(device)
    optimizer = torch.optim.AdamW(net.parameters(), lr=1e-3, weight_decay=1e-2)

    print("======= Training Start =======")

    with alive_bar(args.train_epoch, enrich_print=False) as bar:
        loss_rec_his = loss_vq_his = 1e7
        stable_epochs = 0
        if args.codebook_init == "Random":
            warm_epochs = 0
        else:
            # For Kmeans and Geometric initialization
            warm_epochs = min(50, int(args.train_epoch * 0.2))
        for epoch in range(args.train_epoch):
            bar()
            if epoch < warm_epochs:
                warm_one_epoch(
                    model=net,
                    data_types=args.data_type,
                    dataloader=dataloader_train,
                    optimizer=optimizer,
                    epoch=epoch,
                    device=device,
                )
            elif epoch == warm_epochs:
                embeds, ids, _, _, _ = inference(
                    model=net,
                    data_types=args.data_type,
                    data_loader=dataloader_eval,
                    device=device,
                )
                net.quantizer.init_codebook(embeds, method=args.codebook_init)
                if omics_num == 1:
                    net.copy_decoder_q()
            else:
                loss_rec, loss_vq = train_one_epoch(
                    model=net,
                    data_types=args.data_type,
                    dataloader=dataloader_train,
                    optimizer=optimizer,
                    epoch=epoch,
                    device=device,
                )
                converge = (abs(loss_vq_his - loss_vq) <= 1e-5) and (
                    abs(loss_rec_his - loss_rec) <= 1e-5
                )
                if converge:
                    stable_epochs += 1
                    if stable_epochs >= args.converge_threshold:
                        print("Early Stopping.")
                        break
                else:
                    stable_epochs = 0
                    loss_rec_his = loss_rec
                    loss_vq_his = loss_vq

    print("======= Training Done =======")

    print("")

    print("======= Inference Start =======")
    embeds, ids, delta_confs, rec_q_percent, loss_codebook = inference(
        model=net, data_types=args.data_type, data_loader=dataloader_eval, device=device
    )

    print("* Quantized Reconstruction Percent:", rec_q_percent)
    print("* Metacell Delta Assignment Confidence:", np.mean(delta_confs))
    print("* Codebook Loss:", loss_codebook)
    print("")

    if not os.path.exists("./save/"):
        os.makedirs("./save/")

    for i in range(omics_num):
        metacell_path = (
            "./save/"
            + args.save_name
            + "_"
            + args.data_type[i]
            + "_"
            + str(args.metacell_num)
            + "metacell.h5ad"
        )
        adata = adata_list[i]
        metacell_adata = compute_metacell(adata, ids, args)
        metacell_adata.write_h5ad(metacell_path)
        print(args.data_type[i] + " metacell data saved at:", metacell_path)

    assignment_path = (
        "./save/" + args.save_name + "_" + str(args.metacell_num) + "metacell_ids.h5ad"
    )
    adata = sc.AnnData(embeds, dtype=np.float32)
    adata.obs["metacell"] = ids
    if args.type_key in adata_list[0].obs_keys():
        adata.obs[args.type_key] = adata_list[0].obs[args.type_key]
    sc.set_figure_params(figsize=(7, 7), dpi=300)
    sc.pp.neighbors(adata, use_rep="X", metric="cosine")
    sc.tl.umap(adata)
    if args.type_key in adata.obs_keys():
        sc.pl.umap(
            adata,
            color=[args.type_key],
            save="_" + args.save_name + "_embedding.png",
            palette=sns.color_palette(
                "husl", np.unique(adata.obs[args.type_key].values).size
            ),
            show=False,
        )
    rcParams.update(matplotlib.rcParamsDefault)
    adata.write_h5ad(assignment_path)
    print("Metacell assignment saved at:", assignment_path)
    print("")

    fig_save_name = args.save_name + "_" + str(args.metacell_num) + "metacell"
    plot_metacell_umap(adata, fig_save_name)
    plot_metacell_size(adata, fig_save_name)
    if args.type_key in adata.obs_keys():
        plot_celltype_purity(adata, adata.obs[args.type_key], fig_save_name)
    for i in range(omics_num):
        plot_compactness_separation(
            dataloader_train.dataset.raw_list[i].numpy(),
            adata,
            fig_save_name + "_" + args.data_type[i],
            omics_num > 1,
        )

    print("======= Inference Done =======")


def run_metaq(
    data_path,
    data_type,
    save_name,
    metacell_num,
    type_key="celltype",
    codebook_init="Random",
    train_epoch=300,
    batch_size=512,
    converge_threshold=10,
    random_seed=1,
    device="cuda",
):
    args = argparse.Namespace(
        data_path=data_path,
        data_type=data_type,
        save_name=save_name,
        metacell_num=metacell_num,
        type_key=type_key,
        codebook_init=codebook_init,
        train_epoch=train_epoch,
        batch_size=batch_size,
        converge_threshold=converge_threshold,
        random_seed=random_seed,
        device=device,
    )

    assert len(args.data_path) == len(
        args.data_type
    ), "Number of data path and data type mismatch"

    # Randomization
    random.seed(args.random_seed)
    np.random.seed(args.random_seed)
    torch.manual_seed(args.random_seed)
    torch.random.manual_seed(args.random_seed)
    if args.device == "cuda":
        torch.cuda.manual_seed_all(args.random_seed)

    main(args)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # Data configs (User defined)
    parser.add_argument("--data_path", nargs="+", type=str)
    parser.add_argument(
        "--data_type", nargs="+", type=str, choices=["RNA", "ADT", "ATAC"]
    )
    parser.add_argument("--save_name", type=str)
    parser.add_argument("--metacell_num", type=int)
    parser.add_argument("--type_key", type=str, default="celltype")

    # Training configs (Recommend to use the default)
    parser.add_argument(
        "--codebook_init",
        type=str,
        choices=["Random", "Kmeans", "Geometric"],
        default="Random",
    )
    parser.add_argument("--train_epoch", type=int, default=300)
    parser.add_argument("--batch_size", type=int, default=512)
    parser.add_argument("--converge_threshold", type=int, default=10)
    parser.add_argument("--random_seed", type=int, default=1)
    parser.add_argument("--device", type=str, default="cuda")

    args = parser.parse_args()

    assert len(args.data_path) == len(
        args.data_type
    ), "Number of data path and data type mismatch"

    # Randomization
    random.seed(args.random_seed)
    np.random.seed(args.random_seed)
    torch.manual_seed(args.random_seed)
    torch.random.manual_seed(args.random_seed)
    if args.device == "cuda":
        torch.cuda.manual_seed_all(args.random_seed)

    main(args)



import torch
import numpy as np
import scanpy as sc
from scipy import sparse
from torch.utils.data import Dataset, DataLoader


class MetaQDataset(Dataset):
    def __init__(self, x_list, sf_list, raw_list):
        super().__init__()
        self.x_list = x_list
        self.sf_list = sf_list
        self.raw_list = raw_list

        self.cell_num = self.x_list[0].shape[0]
        self.omics_num = len(self.x_list)

        for i in range(self.omics_num):
            self.x_list[i] = torch.from_numpy(self.x_list[i]).float()
            self.sf_list[i] = torch.from_numpy(self.sf_list[i]).float()
            self.raw_list[i] = torch.from_numpy(self.raw_list[i]).float()

    def __len__(self):
        return int(self.cell_num)

    def __getitem__(self, index):
        x_list = []
        sf_list = []
        raw_list = []
        for i in range(self.omics_num):
            x_list.append(self.x_list[i][index])
            sf_list.append(self.sf_list[i][index])
            raw_list.append(self.raw_list[i][index])
        data = {"x": x_list, "sf": sf_list, "raw": raw_list}
        return data


def preprocess(adata, data_type):
    if isinstance(adata.X, sparse.csr_matrix) or isinstance(adata.X, sparse.csc_matrix):
        adata.X = adata.X.toarray()
    raw = adata.X.copy()

    if data_type == "RNA":
        sc.pp.normalize_total(adata, target_sum=1e4)
        sf = np.array((raw.sum(axis=1) / 1e4).tolist()).reshape(-1, 1)
        sc.pp.log1p(adata)
        adata_ = adata.copy()
        if adata.shape[1] < 5000:
            sc.pp.highly_variable_genes(adata, n_top_genes=3000)
        else:
            sc.pp.highly_variable_genes(adata)
        hvg_index = adata.var["highly_variable"].values
        raw = raw[:, hvg_index]
        adata = adata[:, hvg_index]
    elif data_type == "ADT":
        sc.pp.normalize_total(adata, target_sum=1e4)
        sf = np.array((raw.sum(axis=1) / 1e4).tolist()).reshape(-1, 1)
        sc.pp.log1p(adata)
        adata_ = adata.copy()
    elif data_type == "ATAC":
        sc.pp.normalize_total(adata, target_sum=1e4)
        sf = np.array((raw.sum(axis=1) / 1e4).tolist()).reshape(-1, 1)
        sc.pp.log1p(adata)
        adata_ = adata.copy()
        sc.pp.highly_variable_genes(adata, n_top_genes=30000)
        hvg_index = adata.var["highly_variable"].values
        raw = raw[:, hvg_index]
        adata = adata[:, hvg_index]

    sc.pp.scale(adata, max_value=10)
    x = adata.X

    return x, sf, raw, adata_


def load_data(args):
    print("=======Loading and Preprocessing Data=======")

    num_omics = len(args.data_path)
    print("Data of", num_omics, "omics in total")

    x_list = []
    sf_list = []
    raw_list = []
    adata_list = []
    for i in range(num_omics):
        data_path = args.data_path[i]
        data_type = args.data_type[i]

        adata = sc.read_h5ad(data_path)
        x, sf, raw, adata = preprocess(adata, data_type)

        x_list.append(x)
        sf_list.append(sf)
        raw_list.append(raw)
        adata_list.append(adata)

        print(data_path, "loaded with shape", list(x.shape))

    dataset = MetaQDataset(x_list, sf_list, raw_list)
    if args.metacell_num > 1000 and args.batch_size <= 512:
        args.batch_size = 4096
    dataloader_train = DataLoader(
        dataset=dataset,
        batch_size=args.batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=4,
        pin_memory=True,
    )
    dataloader_eval = DataLoader(
        dataset=dataset,
        batch_size=args.batch_size * 4,
        shuffle=False,
        drop_last=False,
        num_workers=4,
    )

    input_dims = [x.shape[1] for x in x_list]

    return adata_list, dataloader_train, dataloader_eval, input_dims


def compute_metacell(adata, meta_ids, args):
    meta_ids = meta_ids.astype(int)
    non_empty_metacell = np.zeros(meta_ids.max() + 1).astype(bool)
    non_empty_metacell[np.unique(meta_ids)] = True

    data = adata.X
    data_meta = np.stack(
        [data[meta_ids == i].mean(axis=0) for i in range(meta_ids.max() + 1)]
    )
    data_meta = data_meta[non_empty_metacell]
    metacell_adata = sc.AnnData(data_meta)

    if args.type_key in adata.obs_keys():
        type_int = torch.from_numpy(adata.obs[args.type_key].cat.codes.values).long()
        type_map = {
            i: adata.obs[args.type_key].cat.categories[i]
            for i in range(type_int.max() + 1)
        }
        type_one_hot = torch.zeros(type_int.shape[0], type_int.max() + 1)
        type_one_hot.scatter_(1, type_int.unsqueeze(1), 1)
        type_meta = (
            torch.stack(
                [
                    type_one_hot[meta_ids == i].mean(dim=0)
                    for i in range(meta_ids.max() + 1)
                ]
            )
            .argmax(dim=1)
            .numpy()
        )
        type_meta = np.array([type_map[i] for i in type_meta])
        type_meta = type_meta[non_empty_metacell]
        metacell_adata.obs[args.type_key] = type_meta

    return metacell_adata



import torch
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import matplotlib.pyplot as plt
from scipy import sparse
from alive_progress import alive_bar
from sklearn.metrics import balanced_accuracy_score


def pearson_pytorch(A, B):
    if isinstance(A, sparse.csr_matrix) or isinstance(A, sparse.csc_matrix):
        A = A.toarray()
    if isinstance(B, sparse.csr_matrix) or isinstance(B, sparse.csc_matrix):
        B = B.toarray()
    A = torch.from_numpy(A).half()
    B = torch.from_numpy(B).half()
    if torch.cuda.is_available():
        A = A.cuda()
        B = B.cuda()

    A_mean = A - A.mean(dim=1, keepdim=True)
    B_mean = B - B.mean(dim=1, keepdim=True)
    A_std = A_mean.norm(dim=1)
    B_std = B_mean.norm(dim=1)
    cov_matrix = torch.mm(A_mean, B_mean.t())
    correlation_matrix = cov_matrix / torch.outer(A_std, B_std)

    return correlation_matrix.cpu().numpy()


def pairwise_correlation(data, save_name):
    cell_num = data.shape[0]
    save_path = "./save/" + save_name + "_" + str(cell_num) + "cells_correlation.npy"
    try:
        corr = np.load(save_path)
        print("Correlation Loaded from:", save_path)
    except:
        print("Computing Pairwise Correlation...")
        corr = np.zeros((cell_num, cell_num), dtype=np.float16)
        chunk_size = 5000
        chunks_num = cell_num // chunk_size + 1 if cell_num % chunk_size != 0 else 0
        total_steps = chunks_num * chunks_num
        with alive_bar(total_steps, enrich_print=False) as bar:
            for i in range(0, cell_num, chunk_size):
                row_start, row_end = i, min(i + chunk_size, cell_num)
                for j in range(0, cell_num, chunk_size):
                    col_start, col_end = j, min(j + chunk_size, cell_num)
                    corr[row_start:row_end, col_start:col_end] = pearson_pytorch(
                        data[row_start:row_end], data[col_start:col_end]
                    )
                    bar()
        np.save(save_path, corr)
        print("Correlation Saved to:", save_path)
    return corr


def compactness(corr, metacell_assign):
    cell_num = corr.shape[0]
    metacell_num = len(np.unique(metacell_assign))
    assignment_ids = np.unique(metacell_assign)
    compactness = []
    for i in assignment_ids:
        idx = np.where(metacell_assign == i)[0]
        if len(idx) == 0:
            continue
        compactness.append(
            np.mean(corr[idx][:, idx]) * len(idx) / cell_num * metacell_num
        )
    return compactness


def separation(corr, metacell_assign):
    cell_num = corr.shape[0]
    metacell_num = len(np.unique(metacell_assign))
    assignment_ids = np.unique(metacell_assign)
    separation = []
    for i in assignment_ids:
        idx = np.where(metacell_assign == i)[0]
        complementary_idx = np.where(metacell_assign != i)[0]
        if len(idx) == 0:
            continue
        separation.append(
            np.mean(1 - corr[idx][:, complementary_idx].max(axis=1))
            * len(idx)
            / cell_num
            * metacell_num
        )
    return separation


def plot_compactness_separation(
    raw_count, assignment_adata, save_name, multi_omics=False
):
    raw_adata = sc.AnnData(raw_count)
    sc.pp.normalize_total(raw_adata, target_sum=1e4)
    sc.pp.log1p(raw_adata)

    if not multi_omics:
        corr_save_name = save_name.split("_")[0]
    else:
        corr_save_name = save_name.split("_")[0] + "_" + save_name.split("_")[-1]
    corr = pairwise_correlation(raw_adata.X, corr_save_name)
    assignments = assignment_adata.obs["metacell"]

    compactness_scores = compactness(corr, assignments)
    separation_scores = separation(corr, assignments)

    np.savetxt("./save/" + save_name + "_compactness.txt", np.array(compactness_scores))
    np.savetxt("./save/" + save_name + "_separation.txt", np.array(separation_scores))

    print("* Average Compactness Score:", np.mean(compactness_scores))
    print("* Average Separation Score:", np.mean(separation_scores))

    data = pd.DataFrame(columns=["Metric", "Score"])
    for score in compactness_scores:
        data = pd.concat(
            [data, pd.DataFrame({"Metric": ["Compactness"], "Score": [score]})]
        )
    for score in separation_scores:
        data = pd.concat(
            [data, pd.DataFrame({"Metric": ["Separation"], "Score": [score]})]
        )

    plt.figure(figsize=(4, 7), dpi=300)
    sns.set_theme(style="ticks", font_scale=1.0)

    sns.boxplot(
        data=data,
        y="Score",
        x="Metric",
        saturation=0.55,
        fliersize=0.5,
        linewidth=0.5,
        width=0.87,
    )

    plt.title("Metacell Metric")
    plt.tight_layout()

    plt.savefig("./figures/" + save_name + "_metric.png", transparent=False)
    plt.close()


def plot_metacell_umap(adata, save_name, meta_size=10, cell_size=0.5):
    umap = (
        pd.DataFrame(adata.obsm["X_umap"])
        .set_index(adata.obs_names)
        .join(adata.obs["metacell"])
    )
    umap["metacell"] = umap["metacell"].astype("category")
    mcs = umap.groupby("metacell").mean().reset_index()

    plt.figure(figsize=(7, 7), dpi=300)
    sns.set_theme(style="ticks", font_scale=1.0)

    sns.scatterplot(x=0, y=1, hue="metacell", data=umap, s=cell_size, legend=None)
    sns.scatterplot(
        x=0,
        y=1,
        s=meta_size,
        hue="metacell",
        data=mcs,
        edgecolor="black",
        linewidth=1,
        legend=None,
    )

    plt.xlabel(f"UMAP1")
    plt.ylabel(f"UMAP2")
    plt.title("Metacell Assignment")
    plt.tight_layout()

    plt.savefig("./figures/" + save_name + "_umap.png", transparent=False)
    plt.close()


def plot_metacell_size(adata, save_name, bins=50):
    label_df = adata.obs["metacell"].reset_index()
    count = label_df.groupby("metacell").count().iloc[:, 0]

    plt.figure(figsize=(7, 3), dpi=300)
    sns.set_theme(style="ticks", font_scale=1.0)

    sns.histplot(data=count, stat="density", bins=bins, kde=True)

    plt.xlabel("Number of Cells per Metacell")
    plt.title("Metacell Size")
    plt.tight_layout()

    plt.savefig("./figures/" + save_name + "_size.png", transparent=False)
    plt.close()


def plot_celltype_purity(adata, annotations, save_name):
    celltypes, class_size = np.unique(annotations, return_counts=True)
    celltype2int = {celltype: i for i, celltype in enumerate(celltypes)}

    assignments = adata.obs["metacell"]
    assignment_ids = np.unique(assignments)

    majority_type_pred = annotations.copy()
    for i in assignment_ids:
        idx = assignments == i
        majority_type = annotations[idx].value_counts().idxmax()
        majority_type_pred[idx] = majority_type

    annotations_int = np.array([celltype2int[celltype] for celltype in annotations])
    majority_type_pred_int = np.array(
        [celltype2int[celltype] for celltype in majority_type_pred]
    )
    balanced_acc = balanced_accuracy_score(annotations_int, majority_type_pred_int)
    print("* Balanced Cell Type Purity:", balanced_acc)

    data = pd.DataFrame(columns=["Cell Type", "Prediction by Majority", "Fraction"])

    sorted_indices = np.argsort(class_size)[::-1]
    sorted_celltypes = celltypes[sorted_indices]

    for type in celltypes:
        type_idx = annotations == type
        preds = majority_type_pred[type_idx]
        for type_ in celltypes:
            total_number = (preds == type_).sum()
            frac = total_number / type_idx.sum()
            data = data.append(
                {
                    "Cell Type": type,
                    "Prediction by Majority": type_,
                    "Fraction": frac,
                },
                ignore_index=True,
            )
    data["Cell Type"] = pd.CategoricalIndex(
        data["Cell Type"], categories=sorted_celltypes
    )
    data.sort_values("Cell Type")
    data["Prediction by Majority"] = pd.CategoricalIndex(
        data["Prediction by Majority"],
        categories=sorted_celltypes,
    )
    data.sort_values("Prediction by Majority")
    data = data.pivot("Cell Type", "Prediction by Majority", "Fraction")
    data = data.fillna(0)

    plt.figure(figsize=(7.3, 7), dpi=300)
    sns.set_theme(style="ticks", font_scale=1.0)

    ax = sns.heatmap(
        data=data,
        vmin=0,
        vmax=1,
        cmap=sns.color_palette("Reds", as_cmap=True),
        cbar=False,
    )
    ax.set_xticklabels(ax.get_xticklabels(), rotation=30, horizontalalignment="right")

    plt.title("Balanced Cell Type Purity: {:.2f}%".format(balanced_acc * 100))
    plt.tight_layout()

    plt.savefig("./figures/" + save_name + "_purity_heatmap.png", transparent=False)
    plt.close()

    data = pd.DataFrame(columns=["Cell Type", "Metacell Purity"])
    purities = []
    for i in assignment_ids:
        idx = assignments == i
        majority_type = annotations[idx].value_counts().idxmax()
        purity = np.sum(annotations[idx] == majority_type) / len(annotations[idx])
        purities.append(purity)
        data = data.append(
            {
                "Cell Type": majority_type,
                "Metacell Purity": purity,
            },
            ignore_index=True,
        )
    avg_purity = data["Metacell Purity"].mean()
    print("* Average Cell Type Purity:", avg_purity)
    np.savetxt("./save/" + save_name + "_purity.txt", np.array(purities))

    plt.figure(figsize=(7, 4), dpi=300)

    ax = sns.boxplot(
        data=data,
        x="Cell Type",
        y="Metacell Purity",
        saturation=0.55,
        fliersize=0.5,
        linewidth=0.5,
        width=0.87,
    )
    ax.set_xticklabels(ax.get_xticklabels(), rotation=30, horizontalalignment="right")

    plt.title("Metacell Purity")
    plt.tight_layout()

    plt.savefig("./figures/" + save_name + "_purity_box.png", transparent=False)
    plt.close()
